/*
 *  linux/include/asm-arm/proc-armo/uncompress.h
 *
 *  Copyright (C) 1997 Russell King
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#define proc_decomp_setup()
