/*
 * i2c bus addresses for the chips supported by tvaudio.c
 */

#define I2C_TDA8425        0x82
#define I2C_TDA9840        0x84
#define I2C_TDA985x_L      0xb4 /* also used by 9873 */
#define I2C_TDA985x_H      0xb6

#define I2C_TEA6300        0x80
#define I2C_TEA6420	   0x98

#define I2C_PIC16C54       0x96 /* PV951 */
