/*
 *  include/asm-s390/shmparam.h
 *
 *  S390 version
 *
 *  Derived from "include/asm-i386/shmparam.h"
 */
#ifndef _ASM_S390_SHMPARAM_H
#define _ASM_S390_SHMPARAM_H

#define SHMLBA PAGE_SIZE                 /* attach addr a multiple of this */

#endif /* _ASM_S390_SHMPARAM_H */
