/* $Id: hdreg.h,v 1.1 2000/01/21 04:56:27 zaitcev Exp $
 * hdreg.h: SPARC PCI specific IDE glue.
 *
 * Copyright (C) 1997  David S. Miller (davem@caip.rutgers.edu)
 * Copyright (C) 1998  Eddie C. Dost   (ecd@skynet.be)
 */

#ifndef __SPARC_HDREG_H
#define __SPARC_HDREG_H

typedef unsigned int ide_ioreg_t;

#endif /* __SPARC_HDREG_H */
