/* $Id: shmparam.h,v 1.4 1999/12/09 10:32:44 davem Exp $ */
#ifndef _ASMSPARC64_SHMPARAM_H
#define _ASMSPARC64_SHMPARAM_H

#define	SHMLBA (PAGE_SIZE<<1)		/* attach addr a multiple of this */

#endif /* _ASMSPARC64_SHMPARAM_H */
